""" Module release async runner for concurrent coroutines processing. """

import asyncio
from typing import Callable, Coroutine, Iterable, Optional

import aiojobs


class AsyncRunner(aiojobs.Scheduler):
    """
    Class for async running coroutines.

    >>> import time
    >>> async def run(): await AsyncRunner(5).runWait([asyncio.sleep(1) for _ in range(10)], True)
    >>> start = time.time()
    >>> asyncio.run(run())
    >>> round(time.time() - start)
    2
    """

    def __init__(
        self,
        concurrency: int,
        exceptionHandler: Optional[Callable[[aiojobs.Scheduler, dict], None]] = None,
        closeTimeout: Optional[int] = None,
    ):
        """
        Args:
            concurrency: coroutines' execution concurrency
            exceptionHandler: function to run on exception occurrence
                should accept two arguments:
                    1. self: AsyncRunner instance
                    2. context: dict having the following format:
                        - message: error message, str
                        - job: failed job, aiojobs._job.Job instance
                        - exception: caught exception, Exception instance
            closeTimeout: timeout for closing runner, None means infinite timeout
        """
        super().__init__(
            close_timeout=closeTimeout,
            limit=concurrency,
            pending_limit=0,
            exception_handler=exceptionHandler,
        )

    def runNoWait(self, coroutines: Iterable[Coroutine], close: bool = False) -> None:
        """
        Start running coroutines.

        Args:
            coroutines: coroutines to run
            close: whether to close AsyncRunner after coroutines are done
        """
        asyncio.create_task(self.runWait(coroutines, close))

    async def runWait(self, coroutines: Iterable[Coroutine], close: bool = True, returnExceptions: bool = True) -> None:
        """
        Run and wait for coroutines.

        Args:
            coroutines: coroutines to run
            close: whether to close AsyncRunner after coroutines are done
            returnExceptions: whether to treat exceptions as results or raise them
        """
        jobs = await asyncio.gather(*map(self.spawn, coroutines))
        await asyncio.gather(
            *(job._do_wait(None) for job in jobs), return_exceptions=returnExceptions  #: pylint: disable-msg=W0212
        )
        if close:
            await self.close()

    def _done(self, job):
        """
        "Correct" handle done job & start the new.
        """
        self._jobs.discard(job)
        if not self.pending_count:
            return
        # No pending jobs when limit is None
        # Safe to subtract.
        ntodo = self._limit - self.active_count
        i = 0
        while i < ntodo:
            if not self.pending_count:
                return
            new_job = self._pending.get_nowait()  #: pylint: disable-msg=C0103
            if new_job.closed or new_job._started.cancelled():  #: pylint: disable-msg=W0212
                continue
            new_job._start()  #: pylint: disable-msg=W0212
            i += 1

    async def close(self):
        """
        "Correct" close runner.
        """
        if self._closed:
            return
        self._closed = True  # prevent adding new jobs

        jobs = self._jobs
        if jobs:
            # cleanup pending queue
            # all job will be started on closing
            while not self._pending.empty():
                self._pending.get_nowait()
            await asyncio.gather(
                *[job._close(self._close_timeout) for job in jobs],  # pylint: disable-msg=protected-access
            )
            self._jobs.clear()
        if not self._failed_task.done():
            self._failed_tasks.put_nowait(None)
            await self._failed_task
        else:
            # case of canceled all tasks in loop
            pass
