"""
Pydantic models
"""

# todo remove pylint: disable-msg=fixme
from datetime import datetime
from enum import Enum
from typing import Any, Callable, List, Optional, Type, TypeVar, Union
from uuid import UUID

import pydantic
from pydantic import AfterValidator, ConfigDict
from pydantic_core import PydanticCustomError
from stringcase import snakecase  # pylint: disable-msg=wrong-import-position
from typing_extensions import Annotated, Literal  # pylint: disable-msg=wrong-import-position
from vlutils.helpers import convertTimeToString  # pylint: disable-msg=wrong-import-position

# TODO: remove after LUNA-6371 completed


T = TypeVar("T")  # pylint: disable-msg=invalid-name


SUFFIX_MAP = {"_gte": "__gte", "_lte": "__lte", "_gt": "__gt", "_lt": "__lt"}


def replaceEqualSuffix(camelCaseName: str) -> str:  # pylint: disable-msg=C0103
    """
    Replace gt/gte/lt/lte suffix contains one underline with two underlines
    Args:
        camelCaseName: key in camelCase
    Returns:
        snake-case with replaced suffix
    """
    snakeCaseName = snakecase(camelCaseName)
    for oneUnderLineKey, twoUnderLineKey in SUFFIX_MAP.items():
        if snakeCaseName.endswith(oneUnderLineKey):
            snakeCaseName = snakeCaseName.replace(oneUnderLineKey, twoUnderLineKey)
            break
    return snakeCaseName


def alias_generator(paramName: str) -> str:  # pylint: disable-msg=C0103
    """
    Generate aliases
    Args:
        paramName: model field name
    """
    return replaceEqualSuffix(snakecase(paramName))


class BaseModel(pydantic.BaseModel):
    """Base model."""

    model_config = ConfigDict(  # type: ignore
        extra="forbid", alias_generator=alias_generator, arbitrary_types_allowed=True, inUTC=True
    )

    # todo remove
    def asDict(self) -> dict:
        """
        Get data from initialized model
        Returns:
            dict with snake_case keys
        """

        def getValue(value_: Any) -> Any:  # pylint: disable-msg=C0103
            """Convert uuid, datetime to string"""
            if isinstance(value_, UUID):
                return str(value_)
            if isinstance(value_, datetime):
                return convertTimeToString(value_, self.model_config.get("inUTC"))  # type: ignore
            if isinstance(value_, Enum):
                return value_.value
            return value_

        res = {}
        for key, value in self.__dict__.items():
            if value is None:
                continue
            snakeCaseKey = replaceEqualSuffix(snakecase(key))
            if hasattr(value, "asDict"):
                res[snakeCaseKey] = value.asDict()
            elif isinstance(value, (list, set)):
                res[snakeCaseKey] = [row.asDict() if isinstance(row, BaseModel) else getValue(row) for row in value]
            else:
                res[snakeCaseKey] = getValue(value)

        return res


class SchemaUpdaterMixinBase:
    """Extended base schema model."""

    # list of inherited classes from the current one
    __inheritors__: list = []

    def __init_subclass__(cls, **kwargs):  # pylint: disable-msg=unused-argument
        cls.__inheritors__.append(cls)

    @classmethod
    def update(cls):
        """Update inherit schema."""

    @classmethod
    def updateSchemas(cls):
        """Run settings update for inherited schemas."""

        for obj in cls.__inheritors__:
            obj.update()


def UpdatableList(  # pylint: disable-msg=invalid-name
    itemType: Type[T],
    minItems: Optional[int] = None,
    maxItems: Optional[int] = None,
    minItemsGetter: Optional[Callable[..., int]] = None,
    maxItemsGetter: Optional[Callable[..., int]] = None,
) -> Type[List[T]]:
    """
    Create pydantic constrained list for schema validation. Update list constraints after creating a service app.

    Args:
        itemType: list item type
        minItems: set min items permanently
        maxItems: set max items permanently
        minItemsGetter: set min items from callable object
        maxItemsGetter: set max items from callable object

    Returns:
        Annotated list
    """

    class UserConstrainedList(SchemaUpdaterMixinBase):
        """User constrained list"""

        __args__ = (itemType,)
        min_items = None
        max_items = None

        @classmethod
        def update(cls):
            """Update"""
            cls.min_items = minItems if minItemsGetter is None else minItemsGetter()
            cls.max_items = maxItems if maxItemsGetter is None else maxItemsGetter()

        @classmethod
        def lenConstraintValidation(cls, value):
            """Constraints workaround that ensures list doesn't reach len limits"""
            if cls.max_items and cls.max_items < len(value):
                raise PydanticCustomError(
                    "ValidationError",
                    f"ensure this value has at most {cls.max_items} items",
                    {"limit_value": cls.max_items},
                )

            if cls.min_items and cls.min_items > len(value):
                raise PydanticCustomError(
                    "ValidationError",
                    f"ensure this value has at least {cls.min_items} items",
                    {"limit_value": cls.min_items},
                )
            return value

    return Annotated[List[itemType], AfterValidator(UserConstrainedList.lenConstraintValidation)]  # type: ignore


def DiscriminatedUnion(key: str, schemas: list[Type[BaseModel]]) -> Type[pydantic.RootModel]:
    # pylint: disable-msg=invalid-name
    """
    Discriminated union factory.
    Args:
        key: discriminator field name
        schemas: schemas
    """
    mapping = {schema.model_fields[key].annotation.__args__[0]: schema for schema in schemas}  # type: ignore
    DiscriminatorModel: Type[BaseModel] = pydantic.create_model(  # type: ignore
        "DiscriminatorModel", **{key: (Literal[*mapping.keys()], ...)}
    )

    class _DiscriminatedUnion(pydantic.RootModel):
        # pylint: disable-msg=missing-function-docstring
        """
        For a valid objects works like a Union.
        For invalid objects calls the additional validation according to the mapping in order to raise proper error.
        """

        root: Union[*mapping.values()]  # type: ignore

        def __init__(self, **kwargs):
            try:
                super().__init__(**kwargs)
            except pydantic.ValidationError:
                badData = kwargs
                DiscriminatorModel.model_validate(badData)
                mapping[badData[key]].model_validate(badData)
                raise

        def __getattr__(self, item: str):
            return getattr(self.root, item)

        def __setattr__(self, item: str, value: Any):
            return setattr(self.root, item, value)

        def asDict(self) -> dict:
            return self.root.asDict()

    return _DiscriminatedUnion
